@@ -8,11 +8,11 @@ Created on Wed Oct 31 16:23:13 2018
 import os
 
 __author__ = 'Sam Schott'
-__version__ = '0.6.4'
+__version__ = '1.0.0'
 __url__ = 'https://github.com/SamSchott/maestral'
 
-# add "~/.local/share" to XDG_DATA_DIRS
+# add '~/.local/share' to XDG_DATA_DIRS
 # this is needed to find icons installed for the current user only
-XDG_DATA_DIRS = os.environ.get("XDG_DATA_DIRS")
-if XDG_DATA_DIRS:
-    os.environ["XDG_DATA_DIRS"] = XDG_DATA_DIRS + ":" + os.path.expanduser("~/.local/share")
+XDG_DATA_DIRS = os.environ.get('XDG_DATA_DIRS', '')
+XDG_DATA_DIRS += ':' + os.path.expanduser('~/.local/share')
+os.environ['XDG_DATA_DIRS'] = XDG_DATA_DIRS.strip(':')
